package com.jerzozwierz

import groovy.json.JsonBuilder
import groovy.json.JsonOutput
import groovy.json.JsonSlurper
import org.apache.commons.io.IOUtils
import org.apache.commons.lang3.StringUtils
import org.springframework.core.io.ClassPathResource

import java.nio.charset.StandardCharsets

class Main {

    private static String RESOURCE_DIR = "src" + File.separator + "main" + File.separator + "resources" + File.separator
    private static String FEAT_DIR = "feats"
    private static String ACTIVATABLE_DIR = "activatableabilities"
    private static String ABILITY_DIR = "abilities"
    private static String SPELL_DIR = "spells"
    private static String ADJUSTED_FILE_PREFIX = "AAAB_adjusted_"


    public static void main(String[] args) {
        Map<PartyMembers, String> partyGUIDMap = new EnumMap<>(PartyMembers.class);
        partyGUIDMap.put(PartyMembers.VALERIE, "5b972eef-5eab-4947-98c5-65d448dc542d")//54be53f0b35bf3c4592a97ae335fe765
        partyGUIDMap.put(PartyMembers.REGONAR, "5dfa8214-7918-4653-a695-368d45e8caa6")//b090918d7e9010a45b96465de7a104c3
        partyGUIDMap.put(PartyMembers.OCTAVIA, "8887ccf5-0832-4356-94a0-ffc850b4ad48")//f9161aa0b3f519c47acbce01f53ee217
        partyGUIDMap.put(PartyMembers.TRISTIAN, "c9bd7d6b-7ec0-43ce-b5f8-f4c0b65450b5")
//f6c23e93512e1b54dba11560446a9e02
        partyGUIDMap.put(PartyMembers.JUBILOST, "67461131-bc56-4809-9480-e9b7f85d2821")
//3f5777b51d301524c9b912812955ee1e
        partyGUIDMap.put(PartyMembers.EKUNDAYO, "904d6ea0-e869-4c5a-b892-85d80ebd5855")
//d5bc1d94cd3e5be4bbc03f3366f67afc
        partyGUIDMap.put(PartyMembers.NOK_NOK, "54413057-dbbb-44f6-a791-1080ff9a4148")//f9417988783876044b76f918f8636455

        //String name, String prefab, String casterReference, String spellbook

        Map<PartyMembers, Character> partyMembersCharacterMap = new LinkedHashMap<>()
        partyMembersCharacterMap.put(PartyMembers.PROTAGONIST,
                new Character("Sarnum", "", "1016b56d-b86b-447f-88d0-8691a9c985a2", "cc2052732997b654e93eac268a39a0a9"))
        partyMembersCharacterMap.put(PartyMembers.LINZI,
                new Character("Linzi", "77c11edb92ce0fd408ad96b40fd27121", "69955af7-b422-4342-ba09-08e85f7faeb7", "bc04fc157a8801d41b877ad0d9af03dd"))
        partyMembersCharacterMap.put(PartyMembers.JAETHAL,
                new Character("Jaethal", "32d2801eddf236b499d42e4a7d34de23", "afa81ad1-fc1d-40da-ab28-40cb17ee9933", "57fab75111f377248810ece84193a5a5"))
        partyMembersCharacterMap.put(PartyMembers.HARRIM,
                new Character("Harrim", "aab03d0ab5262da498b32daa6a99b507", "119140fc-5b69-455c-b524-aa927876dbea", "4673d19a0cf2fab4f885cc4d1353da33"))
        partyMembersCharacterMap.put(PartyMembers.VALERIE,
                new Character("Valerie", "54be53f0b35bf3c4592a97ae335fe765", "a20956c8-d484-47a1-a389-fd6ff8766c32", ""))
        partyMembersCharacterMap.put(PartyMembers.AMIRI,
                new Character("Amiri", "b3f29faef0a82b941af04f08ceb47fa2", "b900549a-d3ca-4533-958e-a509f05a642c", ""))
        Character character = partyMembersCharacterMap.get(PartyMembers.VALERIE)
        int highestId = 6484
        highestId += 2000 //just to be safe
        int ownerIdentifier = Integer.valueOf("3699") //varies from save to save :(
        String ownerGUID = character.getCasterReference()
        String ownerSpellbook = character.getSpellbook()

        def level = "lv2"
        def featFile = "combat_feats_" + level + ".json"
        def activatableFile = "activatable_abilities_" + level + ".json"
        def abilityFile = "abilities_" + level + ".json"
        def spellFile = "spells_" + level + ".json"

        //update feats
        def modifiedFeatText
        def modifiedActivatablesText
        def modifiedAbilitiesText
        def modifiedSpellText
        (highestId, modifiedFeatText) = updateFeats(featFile, highestId, ownerIdentifier, ownerGUID)

        //update activatable abilities
        (highestId, modifiedActivatablesText) = updateActivatables(activatableFile, highestId, ownerIdentifier)

        //update abilities
        (highestId, modifiedAbilitiesText) = updateAbilities(abilityFile, highestId, ownerIdentifier)

        //Link Abilities to feats
        modifiedFeatText = linkAbilitiesToFeats(modifiedAbilitiesText, modifiedFeatText)

        //Link Activatable Abilities to feats
        modifiedFeatText = linkActivatableAbilitiesToFeats(modifiedActivatablesText, modifiedFeatText)

        //update spells
        (highestId, modifiedSpellText) = updateSpells(spellFile, highestId, ownerIdentifier, ownerSpellbook)

        saveFile(RESOURCE_DIR + FEAT_DIR, ADJUSTED_FILE_PREFIX + featFile, modifiedFeatText)
        saveFile(RESOURCE_DIR + ACTIVATABLE_DIR, ADJUSTED_FILE_PREFIX + activatableFile, modifiedActivatablesText)
        saveFile(RESOURCE_DIR + ABILITY_DIR, ADJUSTED_FILE_PREFIX + abilityFile, modifiedAbilitiesText)
        saveFile(RESOURCE_DIR + SPELL_DIR, ADJUSTED_FILE_PREFIX + spellFile, modifiedSpellText)

    }

    static def updateSpells(def inputFile, int highestId, int ownerIdentifier, String spellBook) {
        String payloadPath = SPELL_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidSpellConstants(modifiedFile, ownerIdentifier, spellBook)
        return [lastIdentifier, modifiedFile]
    }

    static def setValidSpellConstants(def textFile, int ownerIdentifier, String spellBook) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Caster.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.m_SpellbookBlueprint = spellBook.toString() }
        return builder.toPrettyString()
    }

    static def linkActivatableAbilitiesToFeats(def activatableText, def modifiedFeatText) {
        def slurpedActivatables = new JsonSlurper().parseText(activatableText)
        def builderActivatables = new JsonBuilder(slurpedActivatables)

        def slurpedFeats = new JsonSlurper().parseText(modifiedFeatText)
        def builderFeats = new JsonBuilder(slurpedFeats)



        builderActivatables.content.each { activatable ->

            def valueIndex = 0
            if (activatable._referenceIndex != null) {
                valueIndex = activatable._referenceIndex
            }

            def componentsDataIndex = 0
            if (activatable._componentsDataIndex != null) {
                componentsDataIndex = activatable._componentsDataIndex
            }

            def identifier = activatable.$id
            def comment = activatable._comment
            println "id: " + identifier + " comment we'll be looking for: " + comment
            builderFeats.content.each { feat ->
                def featComment = feat._comment
                if (StringUtils.contains(featComment, comment)) {
                    println "It seems that we've found a feat that should be linked to a activatable: " + featComment
                    def componentsData = feat.m_ComponentsData
                    componentsData.eachWithIndex { componentData, index ->
                        if (componentData.m_AppliedFacts && index == componentsDataIndex) {
                            def appliedFacts = componentData.m_AppliedFacts
                            appliedFacts.each { appliedFact ->
                                if (appliedFact) {
                                    def values = appliedFact.find { it.key == "\$values" }?.value
                                    if (values) {
                                        values.eachWithIndex { value, idx ->
                                            if (idx == valueIndex) {
                                                value.$ref = identifier
                                            }
                                        }
                                    }
                                }

                            }
                        }
                    }
                }
            }
        }

        return builderFeats.toPrettyString()
    }

    static def linkAbilitiesToFeats(def abilityText, def modifiedFeatText) {
        def slurpedAbilities = new JsonSlurper().parseText(abilityText)
        def builderAbilities = new JsonBuilder(slurpedAbilities)

        def slurpedFeats = new JsonSlurper().parseText(modifiedFeatText)
        def builderFeats = new JsonBuilder(slurpedFeats)

        builderAbilities.content.each { ability ->
            def identifier = ability.$id
            def comment = ability._comment
            def valueIndex = 0
            if (ability._referenceIndex != null) {
                valueIndex = ability._referenceIndex
            }

            def componentsDataIndex = 0
            if (ability._componentsDataIndex != null) {
                componentsDataIndex = ability._componentsDataIndex
            }
            println "id: " + identifier + " comment we'll be looking for: " + comment
            builderFeats.content.each { feat ->
                def featComment = feat._comment
                if (StringUtils.contains(featComment, comment)) {
                    println "It seems that we've found a feat that should be linked to a ability: " + featComment
                    def componentsData = feat.m_ComponentsData
                    componentsData.eachWithIndex { componentData, index ->
                        if (componentData.m_AppliedFacts && index == componentsDataIndex) {
                            def appliedFacts = componentData.m_AppliedFacts
                            appliedFacts.each { appliedFact ->
                                if (appliedFact) {
                                    def values = appliedFact.find { it.key == "\$values" }?.value
                                    if (values) {
                                        values.eachWithIndex { value, idx ->
                                            if (idx == valueIndex) {
                                                value.$ref = identifier
                                            }
                                        }
                                    }
                                }

                            }
                        }
                    }
                }
            }
        }

        return builderFeats.toPrettyString()
    }

    static def updateFeats(def inputFile, int highestId, int ownerIdentifier, String ownerGUID) {
        String payloadPath = FEAT_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidFeatConstants(modifiedFile, ownerIdentifier, ownerGUID)
        return [lastIdentifier, modifiedFile]
    }

    static def updateActivatables(def inputFile, int highestId, int ownerIdentifier) {
        String payloadPath = ACTIVATABLE_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidActivatablesConstants(modifiedFile, ownerIdentifier)
        return [lastIdentifier, modifiedFile]
    }

    static def updateAbilities(def inputFile, int highestId, int ownerIdentifier) {
        String payloadPath = ABILITY_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidAbilityConstants(modifiedFile, ownerIdentifier)
        return [lastIdentifier, modifiedFile]
    }

    static def setValidActivatablesConstants(def textFile, int ownerIdentifier) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Owner.$ref = ownerIdentifier.toString() }
        return builder.toPrettyString()
    }

    static def setValidAbilityConstants(def textFile, int ownerIdentifier) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Owner.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.Data.Caster.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.Data.Fact.$ref = it.$id.toString() }
        return builder.toPrettyString()
    }

    static def incrementIdentifiers(def textFile, int identifier) {
        def lines = textFile.readLines()
        def newLines = ""
        String identifierPattern = ".*\\\$id\": \"\\d+\","
        for (def line in lines) {
            String newLine
            if (line.matches(identifierPattern)) {
                identifier++
                newLine = line.replaceAll("\\d+", identifier.toString())
                newLines += newLine
                continue
            }
            newLines += line
        }
//        println "updated file: "+beautifyJson(newLines)
        return [newLines, identifier]

    }

    static def setValidFeatConstants(def textFile, int ownerIdentifier, String ownerGUID) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Owner.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.m_Context.m_OwnerDescriptor.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.m_Context.m_CasterReference.m_UniqueId = ownerGUID }
        return builder.toPrettyString()
    }

    static def linkFactsIfNeeded(def textFile) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { feat ->
            def special = feat._special
            if (special && StringUtils.contains(special, "linkFactToAppliedFacts")) {
                def comment = feat._comment
                def componentsData = feat.m_ComponentsData
                componentsData.each { componentData ->
                    if (componentData.m_AppliedFact) {
                        def identifier = componentData.m_AppliedFact.$id
                        println "linking Fact to AppliedFacts for feat " + comment
                        componentData.m_AppliedFact.Data.Fact.$ref = identifier
                    }
                }
            }
        }
        return builder.toPrettyString()
    }

    static String readFileAsText(String directory, String fileName) {

        return IOUtils.toString(new ClassPathResource(directory + fileName).getInputStream(), StandardCharsets.UTF_8)
    }

    static def beautifyJson(String text) {
        return JsonOutput.prettyPrint(text.toString())
    }

    static def saveFile(String directory, String fileName, String text) {
        File file = new File(directory, fileName)
        if(file.exists()){
            file.delete()
            file.createNewFile()
        }
        file.withWriter('utf-8') { writer ->
            writer.write(text)
            writer.close()
        }

    }


}
