package com.pkm

import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.jerzozwierz.Character
import com.jerzozwierz.PartyMembers
import groovy.json.JsonBuilder
import groovy.json.JsonSlurper
import groovy.time.TimeCategory
import groovy.time.TimeDuration
import org.apache.commons.compress.archivers.zip.ZipArchiveEntry
import org.apache.commons.compress.archivers.zip.ZipFile
import org.apache.commons.io.IOUtils
import org.apache.commons.lang3.StringUtils
import org.apache.tools.zip.ZipEntry
import org.springframework.core.io.ClassPathResource

import java.nio.charset.StandardCharsets
import java.nio.file.Files
import java.nio.file.Paths

class SaveEdit {
    private static String RESOURCE_DIR = "src" + File.separator + "main" + File.separator + "resources" + File.separator
    private static String UNZIPPED_SAVE_DIR = "unzippedSave"
    private static String UNZIPPED_SAVE_FILES_PATH = RESOURCE_DIR + UNZIPPED_SAVE_DIR + File.separator
    private static String PARTY_FILE = RESOURCE_DIR + UNZIPPED_SAVE_DIR + File.separator + "party.json"
    private static String TWO_WEAP_IDENTIFIER = "0"

    private static String FEAT_DIR = "feats"
    private static String ACTIVATABLE_DIR = "activatableabilities"
    private static String ABILITY_DIR = "abilities"
    private static String SPELL_DIR = "spells"
    private static String ADJUSTED_FILE_PREFIX = "AAA_adjusted_"

    public static void main(String[] args) {
        def timeStart = new Date()

        String saveFilePath = "C:\\Users\\NEuRO\\AppData\\LocalLow\\Owlcat Games\\Pathfinder Kingmaker\\Saved Games\\"
        Map<PartyMembers, Character> partyMembersCharacterMap = new LinkedHashMap<>()
        partyMembersCharacterMap.put(PartyMembers.PROTAGONIST,
                new Character("Sarnum", ""))

        partyMembersCharacterMap.put(PartyMembers.LINZI,
                new Character("Linzi", "77c11edb92ce0fd408ad96b40fd27121"))

        partyMembersCharacterMap.put(PartyMembers.JAETHAL,
                new Character("Jaethal", "32d2801eddf236b499d42e4a7d34de23"))

        partyMembersCharacterMap.put(PartyMembers.HARRIM,
                new Character("Harrim", "aab03d0ab5262da498b32daa6a99b507"))

        partyMembersCharacterMap.put(PartyMembers.VALERIE,
                new Character("Valerie", "54be53f0b35bf3c4592a97ae335fe765"))

        partyMembersCharacterMap.put(PartyMembers.AMIRI,
                new Character("Amiri", "b3f29faef0a82b941af04f08ceb47fa2"))

//6447d104a2222c14d9c9b8a36e4eb242
        Character character = partyMembersCharacterMap.get(PartyMembers.PROTAGONIST)
        String saveFile = "Manual_10_before_levelup.zks"
//        saveFile = "Manual_4_Oleg_s_Trading_Post___15_Gozran__IV__4710___23_21_55.zks"
        List<String> wantedFiles = Arrays.asList("party.json")

        extractWantedFilesFromSave(saveFilePath + saveFile, wantedFiles)

        String party = new String(Files.readAllBytes(Paths.get(PARTY_FILE)))
        JsonObject jobj = new Gson().fromJson(party, JsonObject.class)
        int id
        id = findHighestIdentifier(jobj, id)

        println "Found highest id: " + id


        def searchValue = character.getPrefab()
        if (StringUtils.isBlank(character.getPrefab())) {
            searchValue = character.getName()
        }
        JsonObject charObject = findCharacter(jobj, searchValue)
        def charString = charObject.toString()
        int ownerIdentifier = findOwnerId(charString)
        println "Owner Identifier "+ownerIdentifier

        String twoWeapFeatGUID = "6948b379c0562714d9f6d58ccbfa8faa"
        int twoWeapIdentifier = findTwoWeapId(charString, twoWeapFeatGUID)

        TWO_WEAP_IDENTIFIER = twoWeapIdentifier.toString()

        String ownerGUID = getOwnerGUID(charString)
        character.setCasterReference(ownerGUID)
        println character.getName() + " owner GUID " + ownerGUID
        String ownerSpellbook = getOwnerSpellbook(charString)
        character.setSpellbook(ownerSpellbook)

        int highestId = id + 2000 //just to be safe

        def level = "lv3"
        def featFile = "combat_feats_" + level + ".json"
        def activatableFile = "activatable_abilities_" + level + ".json"
        def abilityFile = "abilities_" + level + ".json"
        def spellFile = "spells_" + level + ".json"

        //update feats
        def modifiedFeatText
        def modifiedActivatablesText
        def modifiedAbilitiesText
        def modifiedSpellText


        (highestId, modifiedFeatText) = updateFeats(featFile, highestId, ownerIdentifier, ownerGUID)

        //update activatable abilities
        (highestId, modifiedActivatablesText) = updateActivatables(activatableFile, highestId, ownerIdentifier)

        //update abilities
        (highestId, modifiedAbilitiesText) = updateAbilities(abilityFile, highestId, ownerIdentifier)

        //Link Abilities to feats
        modifiedFeatText = linkAbilitiesToFeats(modifiedAbilitiesText, modifiedFeatText)

        //Link Activatable Abilities to feats
        modifiedFeatText = linkActivatableAbilitiesToFeats(modifiedActivatablesText, modifiedFeatText)

        //update spells
        (highestId, modifiedSpellText) = updateSpells(spellFile, highestId, ownerIdentifier, ownerSpellbook)

        //so we have our file, modded feats
        //update save
        charString = updateCharString(charString,modifiedFeatText,modifiedActivatablesText,modifiedAbilitiesText)

        party = party.replace(charObject.toString(),charString)

        saveToFile(RESOURCE_DIR + FEAT_DIR, ADJUSTED_FILE_PREFIX + featFile, modifiedFeatText)
        saveToFile(RESOURCE_DIR + ACTIVATABLE_DIR, ADJUSTED_FILE_PREFIX + activatableFile, modifiedActivatablesText)
        saveToFile(RESOURCE_DIR + ABILITY_DIR, ADJUSTED_FILE_PREFIX + abilityFile, modifiedAbilitiesText)
        saveToFile(RESOURCE_DIR + SPELL_DIR, ADJUSTED_FILE_PREFIX + spellFile, modifiedSpellText)

        saveToFile(RESOURCE_DIR + UNZIPPED_SAVE_DIR,"party.json",party)

        println character.toString()
        def timeStop = new Date()
        TimeDuration duration = TimeCategory.minus(timeStop, timeStart)
        println "program took": duration

    }

    static Object removeUnwantedElements(Object object){
        if(object._comment != null){
            object.remove("_comment")
        }
        if(object._referenceIndex != null){
            object.remove("_referenceIndex")
        }
        if(object._componentsDataIndex != null){
            object.remove("_componentsDataIndex")
        }
        return object
    }

    static String updateCharString(String characterString, String modifiedFeatText,def modifiedActivatablesText,def modifiedAbilitiesText){
        def slurpedChar = new JsonSlurper().parseText(characterString)
        def builderChar = new JsonBuilder(slurpedChar)

        //updateFeats
        def slurpedFeats = new JsonSlurper().parseText(modifiedFeatText)
        def builderFeats = new JsonBuilder(slurpedFeats)

        builderFeats.content.each{ feat ->
            feat = removeUnwantedElements(feat)
            //add the feat to our character
            builderChar.content.Progression.Features.m_Facts.add(feat)
        }

        //updateActivatableAbilities
        def slurpedActivatables = new JsonSlurper().parseText(modifiedActivatablesText)
        def builderActivatables = new JsonBuilder(slurpedActivatables)

        builderActivatables.content.each{ activatable ->
            activatable = removeUnwantedElements(activatable)
            //add the feat to our character
            builderChar.content.ActivatableAbilities.m_Facts.add(activatable)
        }

        //updateAbilities
        def slurpedAbilities = new JsonSlurper().parseText(modifiedAbilitiesText)
        def builderAbilities = new JsonBuilder(slurpedAbilities)

        builderAbilities.content.each{ ability ->
            ability = removeUnwantedElements(ability)
            //add the feat to our character
            builderChar.content.Abilities.m_Facts.add(ability)
        }

        return builderChar.toString()
    }

    static def linkActivatableAbilitiesToFeats(def activatableText, def modifiedFeatText) {
        def slurpedActivatables = new JsonSlurper().parseText(activatableText)
        def builderActivatables = new JsonBuilder(slurpedActivatables)

        def slurpedFeats = new JsonSlurper().parseText(modifiedFeatText)
        def builderFeats = new JsonBuilder(slurpedFeats)

        builderActivatables.content.each { activatable ->

            Boolean found = false

            def valueIndex = 0
            if (activatable._referenceIndex != null) {
                valueIndex = activatable._referenceIndex
            }

            def componentsDataIndex = 0
            if (activatable._componentsDataIndex != null) {
                componentsDataIndex = activatable._componentsDataIndex
            }

            def identifier = activatable.$id
            def comment = activatable._comment
//            println "id: " + identifier + " comment we'll be looking for: " + comment
            builderFeats.content.each { feat ->
                def featComment = feat._comment
                if (StringUtils.contains(featComment, comment)) {
//                    println "It seems that we've found a feat that should be linked to a activatable: " + featComment
                    found = true
                    def componentsData = feat.m_ComponentsData
                    componentsData.eachWithIndex { componentData, index ->
                        if (componentData.m_AppliedFacts && index == componentsDataIndex) {
                            def appliedFacts = componentData.m_AppliedFacts
                            appliedFacts.each { appliedFact ->
                                if (appliedFact) {
                                    def values = appliedFact.find { it.key == "\$values" }?.value
                                    if (values) {
                                        values.eachWithIndex { value, idx ->
                                            if (idx == valueIndex) {
                                                value.$ref = identifier
                                            }
                                        }
                                    }
                                }

                            }
                        }
                    }
                }
            }

            if (!found) {
                println "[ERROR] Could not find feat " + comment
            }
        }

        return builderFeats.toPrettyString()
    }

    static def getOwnerSpellbook(String characterText) {
        def slurped = new JsonSlurper().parseText(characterText)
        def spellbooks = slurped.m_Spellbooks.Key
        if (spellbooks && spellbooks.size > 0) {
            return spellbooks
        } else {
            return ""
        }

    }

    static def getOwnerGUID(String characterText) {
        def slurped = new JsonSlurper().parseText(characterText)
        def identifiers = slurped.Progression.Features.m_Facts.m_Context.m_CasterReference.m_UniqueId.unique()
        identifiers.removeAll([null])
        if (identifiers.size() == 1) {
            return identifiers[0]
        }
        return ""
    }

    static def findTwoWeapId(String characterText, String twoWeapFeatGUID) {
        def identifier = 0
        def slurped = new JsonSlurper().parseText(characterText)
        def facts = slurped.Progression.Features.m_Facts
        for (def fact in facts) {
            if (fact.Blueprint && StringUtils.contains(fact.Blueprint, twoWeapFeatGUID)) {
                identifier = Integer.valueOf(fact.$id)
            }

        }
        return identifier
    }

    static def findOwnerId(String characterText) {
        def slurped = new JsonSlurper().parseText(characterText)
        return Integer.valueOf(slurped.$id)
    }

    static def findCharacter(def object, String searchFor, Object found = new JsonObject()) {

        if (object instanceof JsonPrimitive) {
        }
        if (object instanceof JsonArray) {
            if (object.size() > 0) {
                for (def element in object) {
                    found = findCharacter(element, searchFor, found)
                }
            }
        }
        if (object instanceof JsonObject) {
            if (object.members) {
                for (def member in object.members) {
                    if (member.value) {
                        if (member.value instanceof JsonPrimitive) {
                            if (member.key.toString() == "Blueprint" || member.key.toString() == "CustomName") {
                                if (StringUtils.contains(member.value.toString(), searchFor)) {
                                    found = object
                                }
                            }
                        }
                        if (member.value instanceof JsonArray) {
                            if (member.value.size() > 0) {
                                for (def element in member.value) {
                                    found = findCharacter(element, searchFor, found)
                                }
                            }
                        }
                        if (member.value instanceof JsonObject) {
                            found = findCharacter(member.value, searchFor, found)
                        }
                    }

                }
            }
        }
        return found

    }

    static int findHighestIdentifier(def object, int id) {

        if (object instanceof JsonPrimitive) {
        }
        if (object instanceof JsonArray) {
            if (object.size() > 0) {
                for (def element in object) {
                    id = findHighestIdentifier(element, id)
                }
            }
        }
        if (object instanceof JsonObject) {
            if (object.members) {
                for (def member in object.members) {
                    if (member.value) {
                        if (member.value instanceof JsonPrimitive) {
                            if (member.key.toString() == "\$id") {
                                if (member.value.asInt > Integer.valueOf(id)) {
                                    id = member.value.asInt
                                }
                            }
                        }
                        if (member.value instanceof JsonArray) {
                            if (member.value.size() > 0) {
                                for (def element in member.value) {
                                    id = findHighestIdentifier(element, id)
                                }
                            }

                        }
                        if (member.value instanceof JsonObject) {
                            id = findHighestIdentifier(member.value, id)
                        }
                    }

                }
            }
        }

        return id
    }

    static def beautifyWantedFiles(String unzippedFilesDirectory, List<String> wantedFiles) {
        for (def file in wantedFiles) {
            String party = new String(Files.readAllBytes(Paths.get(unzippedFilesDirectory + file)))
            def slurped = new JsonSlurper().parseText(party)
            def builder = new JsonBuilder(slurped)
            String beautifiedParty = builder.toPrettyString()

            File fold = new File(unzippedFilesDirectory + file)
            fold.delete()
            File fnew = new File(unzippedFilesDirectory + file)

            try {
                FileWriter f2 = new FileWriter(fnew, false)
                f2.write(beautifiedParty)
                f2.close()
            } catch (IOException e) {
                e.printStackTrace()
            }
        }
    }

    static extractWantedFilesFromSave(String pathToFile, List<String> wantedFiles) {

        File destDir = new File(RESOURCE_DIR + UNZIPPED_SAVE_DIR)
        if (!destDir.exists()) {
            destDir.mkdir()
        }

        ZipFile zipArchive = new ZipFile(pathToFile)
        byte[] buffer = new byte[4096];
        Enumeration<ZipArchiveEntry> archiveFiles = zipArchive.getEntries()

        while (archiveFiles.hasMoreElements()) {
            ZipArchiveEntry zipArchiveEntry = archiveFiles.nextElement()
            if (wantedFiles.contains(zipArchiveEntry.name)) {
//                System.out.println("Extracting file: " + zipArchiveEntry.name)

                File extfile = new File(RESOURCE_DIR + UNZIPPED_SAVE_DIR + File.separator + zipArchiveEntry.name)
                if(extfile.exists()){
                    extfile.delete()
                    extfile.createNewFile()
                }
                InputStream inputStream = zipArchive.getInputStream(zipArchiveEntry)
                FileOutputStream fileOutputStream = new FileOutputStream(extfile)
                try {
                    int numBytes
                    while ((numBytes = inputStream.read(buffer, 0, buffer.length)) != -1)
                        fileOutputStream.write(buffer, 0, numBytes)
                }
                finally {
                    fileOutputStream.close()
                    inputStream.close()
                }

                File backup = new File(RESOURCE_DIR + UNZIPPED_SAVE_DIR + File.separator + zipArchiveEntry.name+".bak")
                Files.deleteIfExists(backup.toPath());
                backup << extfile.text
            }
        }
        zipArchive.close()
    }

    static unpackSave_bak(String pathToFile, List<String> wantedFiles) {
        File destDir = new File(RESOURCE_DIR + UNZIPPED_SAVE_DIR);
        byte[] buffer = new byte[1024];
//        ZipInputStream zis = new ZipInputStream(new FileInputStream(pathToFile));
//        ZipEntry zipEntry = zis.getNextEntry();
//        while (zipEntry != null) {
//            if (!wantedFiles.contains(zipEntry.name)) {
//                continue
//            }
//
//
//            File newFile = newFile(destDir, zipEntry);
//            FileOutputStream fos = new FileOutputStream(newFile);
//            int len;
//            while ((len = zis.read(buffer)) > 0) {
//                fos.write(buffer, 0, len);
//            }
//            fos.close();
//            zipEntry = zis.getNextEntry();
//        }
//        zis.closeEntry();
//        zis.close();
    }

    static File newFile(File destinationDir, ZipEntry zipEntry) throws IOException {
        File destFile = new File(destinationDir, zipEntry.getName())

        String destDirPath = destinationDir.getCanonicalPath()
        File directory = new File(destDirPath)
        if (!directory.exists()) {
            directory.mkdir()
        }

        String destFilePath = destFile.getCanonicalPath()

        if (!destFilePath.startsWith(destDirPath + File.separator)) {
            throw new IOException("Entry is outside of the target dir: " + zipEntry.getName())
        }

        return destFile
    }

    static def navigateEntireJSONOBject(def object) {
        if (object instanceof JsonPrimitive) {
        }
        if (object instanceof JsonArray) {
            if (object.size() > 0) {
                object.each { def element -> findHighestIdentifier(element) }
            }
        }
        if (object instanceof JsonObject) {
            if (object.members) {
                object.members.each { def member ->
                    if (member.value) {
                        if (member.value instanceof JsonPrimitive) {
                        }
                        if (member.value instanceof JsonArray) {
                            if (member.value.size() > 0) {
                                member.value.each { def element ->
                                    findHighestIdentifier(element)
                                }
                            }

                        }
                        if (member.value instanceof JsonObject) {
                            findHighestIdentifier(member.value)
                        }
                    }

                }
            }
        }
    }


    static def linkAbilitiesToFeats(def abilityText, def modifiedFeatText) {
        def slurpedAbilities = new JsonSlurper().parseText(abilityText)
        def builderAbilities = new JsonBuilder(slurpedAbilities)

        def slurpedFeats = new JsonSlurper().parseText(modifiedFeatText)
        def builderFeats = new JsonBuilder(slurpedFeats)

        builderAbilities.content.each { ability ->

            Boolean found = false

            def identifier = ability.$id
            def comment = ability._comment
            def valueIndex = 0
            if (ability._referenceIndex != null) {
                valueIndex = ability._referenceIndex
            }

            def componentsDataIndex = 0
            if (ability._componentsDataIndex != null) {
                componentsDataIndex = ability._componentsDataIndex
            }
//            println "id: " + identifier + " comment we'll be looking for: " + comment
            builderFeats.content.each { feat ->
                def featComment = feat._comment
                if (StringUtils.contains(featComment, comment)) {
//                    println "It seems that we've found a feat that should be linked to a ability: " + featComment
                    found = true
                    def componentsData = feat.m_ComponentsData
                    componentsData.eachWithIndex { componentData, index ->
                        if (componentData.m_AppliedFacts && index == componentsDataIndex) {
                            def appliedFacts = componentData.m_AppliedFacts
                            appliedFacts.each { appliedFact ->
                                if (appliedFact) {
                                    def values = appliedFact.find { it.key == "\$values" }?.value
                                    if (values) {
                                        values.eachWithIndex { value, idx ->
                                            if (idx == valueIndex) {
                                                value.$ref = identifier
                                            }
                                        }
                                    }
                                }

                            }
                        }
                    }
                }
            }

            if (!found) {
                println "[ERROR] Could not find feat " + comment
            }
        }

        return builderFeats.toPrettyString()
    }

    static def updateFeats(def inputFile, int highestId, int ownerIdentifier, String ownerGUID) {
        String payloadPath = FEAT_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidFeatConstants(modifiedFile, ownerIdentifier, ownerGUID)
        return [lastIdentifier, modifiedFile]
    }

    static def updateActivatables(def inputFile, int highestId, int ownerIdentifier) {
        String payloadPath = ACTIVATABLE_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidActivatablesConstants(modifiedFile, ownerIdentifier)
        return [lastIdentifier, modifiedFile]
    }

    static def updateAbilities(def inputFile, int highestId, int ownerIdentifier) {
        String payloadPath = ABILITY_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidAbilityConstants(modifiedFile, ownerIdentifier)
        return [lastIdentifier, modifiedFile]
    }

    static def setValidActivatablesConstants(def textFile, int ownerIdentifier) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Owner.$ref = ownerIdentifier.toString() }
        return builder.toPrettyString()
    }

    static def setValidAbilityConstants(def textFile, int ownerIdentifier) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Owner.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.Data.Caster.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.Data.Fact.$ref = it.$id.toString() }
        return builder.toPrettyString()
    }

    static def incrementIdentifiers(def textFile, int identifier) {
        def lines = textFile.readLines()
        def newLines = ""
        String identifierPattern = ".*\\\$id\": \"\\d+\","
        for (def line in lines) {
            String newLine
            if (line.matches(identifierPattern)) {
                identifier++
                newLine = line.replaceAll("\\d+", identifier.toString())
                newLines += newLine
                continue
            }
            newLines += line
        }
//        println "updated file: "+beautifyJson(newLines)
        return [newLines, identifier]

    }

    static def setValidFeatConstants(def textFile, int ownerIdentifier, String ownerGUID) {
        def slurped = new JsonSlurper().parseText(textFile)
        for (def comment in slurped._comment) {
            if (StringUtils.contains(comment, "Two-Weap")) {
                println "Remember to update the Rank of Two-Weapon fighting in Your save dummy!!! \"" + TWO_WEAP_IDENTIFIER+"\""
            }
        }
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Owner.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.m_Context.m_OwnerDescriptor.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.m_Context.m_CasterReference.m_UniqueId = ownerGUID }

        //super special handling for two weapon feats, less manual labor
        builder.content.each { it ->
            if (StringUtils.contains(it._comment, "Two-Weap")) {
                def values = it.m_ComponentsData[0].m_AppliedFacts.$values[0].$ref = TWO_WEAP_IDENTIFIER
            }
        }
        return builder.toPrettyString()
    }

    static String readFileAsText(String directory, String fileName) {

        return IOUtils.toString(new ClassPathResource(directory + fileName).getInputStream(), StandardCharsets.UTF_8)
    }

    static def saveToFile(String directory, String fileName, String text) {
        File file = new File(directory, fileName)
        if (!file.exists()) {
            file.createNewFile()
        }
        file.withWriter('utf-8') { writer ->
            writer.write(text)
            writer.close()
        }

    }

    static def updateSpells(def inputFile, int highestId, int ownerIdentifier, String spellBook) {
        String payloadPath = SPELL_DIR + File.separator
        def textFile = readFileAsText(payloadPath, inputFile)
        def (modifiedFile, int lastIdentifier) = incrementIdentifiers(textFile, highestId)
        modifiedFile = setValidSpellConstants(modifiedFile, ownerIdentifier, spellBook)
        return [lastIdentifier, modifiedFile]
    }

    static def setValidSpellConstants(def textFile, int ownerIdentifier, String spellBook) {
        def slurped = new JsonSlurper().parseText(textFile)
        def builder = new JsonBuilder(slurped)
        builder.content.each { it -> it.Caster.$ref = ownerIdentifier.toString() }
        builder.content.each { it -> it.m_SpellbookBlueprint = spellBook.toString() }
        return builder.toPrettyString()
    }

}
